#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3				// Use modern global access method and strict Wave access.
#pragma DefaultTab={3,20,4}		// Set default tab width in Igor Pro 9 and later
#pragma IgorVersion=9.0.0.4		// HDF5DimensionScales and VLEN loading added 
#pragma version=120				// version 1.2.0

//___________________________________________________________________________________________________	
//	QLA netCDF reader - Quacks Like A netCDF
//	A tool for loading netCDF-3 and netCDF-4 files consistently into Igor Pro 
//___________________________________________________________________________________________________	
//
//	Copyright ©2021
//	Chris Reed (FAAM Airborne Laboratory, Building 146, Central Avenue, Cranfield University, Bedfordshire, MK43 0AL, UK)
//	chris.reed@faam.ac.uk
//___________________________________________________________________________________________________	
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//___________________________________________________________________________________________________	

// constant to name all 'time/Time/TIME' waves loaded from netCDF
Static StrConstant timeWvStrConstant = "Time_UTC" 

//#define AlphaNumSort 			// sort attributes alphanumerically, instead of creation order

//___________________________________________________________________________________________________	
// menu
Menu "QLA netCDF" 
	"Load selected",/Q,Print "Load netCDFs: \r\n\t"+ReplaceString(";",QLA_netCDF(2),"\r\n\t")
	"\\M0Load w/o unpacking",/Q,Print "Load netCDFs: \r\n\t"+ReplaceString(";",QLA_netCDF(2,unpack=4),"\r\n\t")
	"Load and transpose",/Q,Print "Load netCDFs: \r\n\t"+ReplaceString(";",QLA_netCDF(2,trans=1),"\r\n\t")
end
//___________________________________________________________________________________________________	
// interactivly load multiple netCDF3/4s or silently load a single netCDF if called from within another function
// options - where data is loaded 
// 		0 - load netCDFs in root
// 		1 - load netCDFs in current data folder 
// 		2 - load netCDFs in folder relative to root 
// 		3 - load netcDFs in folder relative to current data folder 
// unpack - mode for unpacking data
//		0 - default - Range check and data scaling is enabled, fill data is set to NaN ",unpack=0" or ommit
//		1 - Range check and data scaling is enabled ",unpack=1" 
// 		2 - Range check is enabled, data scaling is disabled ",unpack=2"
// 		3 - Range check is disabled, data scaling is enabled ",unpack=3"
// 		4 - Range check and data scaling is disabled ",unpack=4"
// trans - transpose - transpose 2D datasets along the long axis ",trans=1"  
Function/S QLA_netCDF(int DFoption[,int unpack,int transpose])
 
 	// check for the presence of the netCDF xop
 	#if !Exists("NC_OpenFile")
		QLA_error()
	#endif
 
	String savedDF= GetDataFolder(1)
	String fileFilters = "netCDF Data Files (*.nc,*.cdf):.nc,.cdf;HDF5 Data Files (*.h5,*.hdf5):.h5,.hdf5;"
	String fileNameStr,fileLstStr,datafolderStr="",returnStr="",fltNumStr
	Variable i


	Open/D/R/MULT=1/F=fileFilters/M="Select netCDFs to open" fileID
	if(strlen(S_fileName)==0)
		return "" // user didn't choose a file
	else
		fileLstStr=ReplaceString("\r",S_filename,";")
	endif

	
	// get correct path and datafolder name from options
	if (DFoption>1)
		for (i=0;i<ItemsInList(fileLstStr);i+=1)
			if (DFoption==2) // load relative to root
				datafolderStr+="root:"
			elseif (DFoption==3) // load relative to current data folder
				datafolderStr+=savedDF
			endif
			fileNameStr=CleanUpName(ParseFilePath(0,StringFromList(i,fileLstStr),":",1,0),1)
			datafolderStr+=PossiblyQuoteName(ParseFilePath(3, RemoveEnding(fileNameStr,".nc"), ":", 0, 0)[0,254])+";"
		endfor
	endif

	// load files in list 
	for (i=0;i<ItemsInList(fileLstStr);i+=1)
		if (DFoption==0 || DFoption==2)
			SetDataFolder root: // return to root
		elseif (DFoption==1 || DFoption==3)
			SetDataFolder $savedDF // return to current DF
		endif
		if (DFoption==2 || DFoption==3)
			NewDataFolder/O/S $StringFromList(i,dataFolderStr) // create new DF 
		endif
		returnStr+=QLA_get_nc(stringfromList(i,fileLstStr),unpack,transpose)+";" // load the file
	endfor	
	
	SetDataFolder savedDF // return to the start
	
	return returnStr
End
//___________________________________________________________________________________________________	
// function to load netCDF3/4 & hdf5 files, metadata is parsed into the note of the variable. 
// Group and global attribute get ethgier own text wave to attach metadata to
// fill value applied, range checking applied optionally with unpack options (0-4), 
// default is 0 - range check and fill value applied, transposition of multidimensional datasets optionally
Function/S QLA_get_nc(String fNameStr,int mode,int transpose)

	Variable fileID,groupID,i,ii
	String saveDF=GetDataFolder(1)
	String tempWvStr,tempPathStr,tempAttWvStr,tmpVarLstStr="",attStr,varStr,groupLst,path
	String HDF5markup="DIMENSION_LIST;DIMENSION_SCALE;CLASS;NAME;REFERENCE_LIST;_nc3_strict;_NCProperties;_Netcdf4Coordinates;_Netcdf4Dimid;" // list of mark-up we dont want
	
	// Make a new symbolic path to the netCDF
	NewPath/O/Q/Z netCDFpath, ParseFilePath(1, fNameStr, ":", 1, 0)
	
	// get file information
	GetFileFolderInfo/Q/Z/P=netCDFpath fNameStr
		
	// make file information
	Make/O/T/N=(0) $"file_attributes"/Wave=file_info
	Note/k file_info, "// file attributes:\r"
	QLA_set_att(file_info,"filename","\""+ParseFilePath(0,S_path,":",1,0)+"\";")
	QLA_set_att(file_info,"path","\""+ParseFilePath(1,S_path,":",1,0)+"\";")
	QLA_set_att(file_info,"type","\"netCDF\";") // placeholder
	QLA_set_att(file_info,"size","\""+num2str(V_logEOF/(1024^2))+" mb\";")
	QLA_set_att(file_info,"created","\""+Secs2Date(V_modificationDate,2)+", "+Secs2Time(V_modificationDate,3)+"\";")
	
	// try to open as netCDF4/HDF-5 as quicker to fail than netCDF-3
	HDF5OpenFile /Z /R /P=netCDFpath fileID as fNameStr 
	if (V_flag==0) // success! a netCDF-4 or HDF-5
		
		// load entire file
		HDF5LoadGroup /O /TRAN=(transpose) /R /IGOR=0 /CONT=2 /OPTS=1 :,fileID, "."
	
		// Get a list of all the groups in the file, insert a holder for the root group
		HDF5ListGroup /F /R=1 /TYPE=1 /ORDR=1 fileID, "/"
		S_HDF5ListGroup+="/.;" ; groupLst = S_HDF5ListGroup
		
		// group metadata loop
		for (i=0;i<ItemsInList(S_HDF5ListGroup);i+=1)
		
			// Get a list of attributes for each group
			HDF5ListAttributes /Type=1 fileID, StringFromList(i,S_HDF5ListGroup)
			
			// sort attributes alphanumerically
			#ifdef AlphaNumSort
				S_HDF5ListAttributes = SortList(S_HDF5ListAttributes,";",16)
			#endif
			
			// get some file type information before the HDF5markup is stripped
			if (GrepString(ParseFilePath(4,fNameStr,":",0,0),"h5"))
				QLA_set_att(file_info,"type","\"HDF5\";")
			elseif (GrepString(S_HDF5ListAttributes,"_nc3_strict")) 
				QLA_set_att(file_info,"type","\"netCDF-4 classic model\";")
			else
				QLA_set_att(file_info,"type","\"netCDF-4 \";")
			endif
			
			//strip out HDF5/transparent netCDF markup
			S_HDF5ListAttributes = RemoveFromList(HDF5markup,S_HDF5ListAttributes,";")
			
			// get the group name to make info
			if (CmpStr(ParseFilePath(0,StringFromList(i,S_HDF5ListGroup),"/",1,0),".")==0)
				tempWvStr="global_attributes"
			else
				tempWvStr="group_attributes"
			endif
			
			// Get the group info Wave/path
			tempPathStr = PossiblyQuotePath(RemoveEnding(saveDF) + ReplaceString("/",RemoveEnding(StringFromList(i,S_HDF5ListGroup),"/."),":")+":"+tempWvStr,":")

			// Make the info Wave 
			Make /O /T /N=(0) $tempPathStr/Wave=infoWv
			
			// make the attributes title
			if (CmpStr(tempWvStr,"group_attributes") == 0)
				Note/k infoWv, "// group attributes:\r"+note(infoWv)
			else
				Note/k infoWv, "// global attributes:\r"+note(infoWv)
			endif
			
			// Open the group containing the info 
			HDF5OpenGroup /Z fileID, StringFromList(i,S_HDF5ListGroup), groupID
			
			// Load each attribute for the group
			for (ii=0;ii<ItemsInList(S_HDF5ListAttributes);ii+=1)
				attStr = StringFromList(ii,S_HDF5ListAttributes)
				HDF5LoadData /O /A=attStr /TYPE=1 /N=$"temp_"+attStr /OPTS=1 /IGOR=0 /Q /Z groupID, StringFromList(i,S_HDF5ListGroup)
				// string
				if ((WaveType($"temp_"+attStr) == 0) && (numpnts($"temp_"+attStr) != 0)) 
					Wave/Z/T	AttStrWv = $"temp_"+StringFromList(ii,S_HDF5ListAttributes)
					QLA_set_att(infoWv,attStr,"\""+AttStrWv[Inf]+"\";")
				// numeric
				elseif((WaveType($"temp_"+attStr) != 0) && (numpnts($"temp_"+attStr) != 0)) 
					Wave/Z	AttVarWv = $"temp_"+attStr
					QLA_set_att(infoWv,attStr,cprintf(AttVarWv)+"; // "+typeToinfo(WaveType(AttVarWv)))
				// empty
				else
					KillWaves/Z $"temp_"+attStr // kill here as never declared
					QLA_set_att(infoWv,attStr,"n/a")
				endif
			KillWaves/Z AttStrWv, AttVarWv
			endfor
			
			// Kill the info Wave if it is empty - probably an HDF5
			if (ItemsInList(QLA_inq_att(infoWv)) == 0)
				KillWaves/Z infoWv
			endif
			
		// end of group metadata loop
		endfor 	
			
		// group dimensions loop
		for (i=0;i<ItemsInList(groupLst);i+=1)
			tempPathStr = PossiblyQuotePath(RemoveEnding(saveDF) + ReplaceString("/",RemoveEnding(StringFromList(i,groupLst),"/."),":")+":dimensions",":")
			
			// make a wave to hold the dimension information
			Make /O /T /N=(0) $tempPathStr/Wave=dimWv
			Note/k dimWv, "dimensions:\r"+note(dimWv)
			
			// list all the variables in a group
			HDF5ListGroup /F /Type=2 /ORDR=1 fileID, StringFromList(i,groupLst)
			
			// loop though the variables to see if they are a dimension
			for (ii=0;ii<ItemsInList(S_HDF5ListGroup);ii+=1)
				tempWvStr = StringFromList(ii,S_HDF5ListGroup) 
				Note/k dimWv,note(dimWv)+isHDF5Dimension(groupID,tempWvStr)
			endfor
			
			// no dimensions defined in this group
			if (ItemsInlist(listDims(dimWv)) == 0)
				KillWaves dimWv
			endif
		endfor // end of group dimension loop
		
		// List all variables in the whole file (r=1)	
		HDF5ListGroup /F /R=1 /TYPE=2 fileID, "/"
		
		// variable metadata loop 
		for (i=0;i<ItemsInList(S_ObjectPaths);i+=1)
			path = StringFromList(i,S_HDF5ListGroup)
			Wave/Z dataWv = $PossiblyQuotePath(StringFromList(i,S_ObjectPaths),":")
			
			// check to see if data was actually loaded (for HDF5 errors)
			if (!WaveExists(dataWv))
				continue
			endif
			
			// Get a list of attributes for each variable - strip out HDF-5 markup
			HDF5ListAttributes /Type=2 fileID,path
			S_HDF5ListAttributes = RemoveFromlist(HDF5markup,S_HDF5ListAttributes,";")
			
			// sort attributes alphanumerically
			#ifdef AlphaNumSort
				S_HDF5ListAttributes = SortList(S_HDF5ListAttributes,";",16)
			#endif
			
			// Attribute name
			if	(CmpStr(ParseFilePath(0,StringFromList(i,S_ObjectPaths),":",1,0),"time0") == 0)		
				tempWvStr = ReplaceString("ime0",ParseFilePath(0,StringFromList(i,S_ObjectPaths),":",1,0),"ime")
			else
				tempWvStr = ParseFilePath(0,StringFromList(i,S_ObjectPaths),":",1,0)
			endif
			
			// kill the wavenote set the short name
			Note/K dataWv ;	QLA_set_att(dataWv,"short_name","\""+tempWvStr+"\";")

			// Load each attribute for the variable
			for (ii=0;ii<ItemsInList(S_HDF5ListAttributes);ii+=1)
				attStr = StringFromList(ii,S_HDF5ListAttributes)
				HDF5LoadData /O /A=attStr /TYPE=2 /OPTS=1 /IGOR=0 /N=$"temp_"+attStr /Q /Z fileID, path
				// string
				if ((WaveType($"temp_"+attStr)==0) && (numpnts($"temp_"+attStr) != 0))
					Wave/Z/T AttStrWv = $"temp_"+attStr
					QLA_set_att(dataWv,attStr,"\""+AttStrWv[Inf]+"\";")
				// numeric
				elseif((WaveType($"temp_"+attStr) != 0) && (numpnts($"temp_"+attStr) != 0)) 
					Wave/Z AttVarWv = $"temp_"+attStr
					QLA_set_att(dataWv,attStr, cprintf(AttVarWv)+"; // "+typeToinfo(WaveType(AttVarWv)))
				// empty
				else
					KillWaves/Z $"temp_"+attStr // kill here as never declared
					QLA_set_att(dataWv,attStr,"n/a") 
				endif
				KillWaves/Z AttStrWv, AttVarWv
			endfor

			// set the dimensional dependencies
			setDependencies(dataWv,getHDF5Dimensions(fileID, path))

			// apply metadata to wave
			unpackNC(dataWv,mode)
		
		// end of variable metadata loop
		endfor 
		
		// strip out the variable names into a list for file info 
		for (ii=0;ii<ItemsInList(S_ObjectPaths);ii+=1)
			tempPathStr = StringFromList(ii,S_ObjectPaths)
			// deal with variables named the same as a dimension - special netCDF-4 only case
			if (stringMatch(StringFromList(ii,S_ObjectPaths),"*_nc4_non_coord_*"))
				Duplicate/O $PossiblyQuotePath(tempPathStr,":") $ReplaceString("_nc4_non_coord_",tempPathStr,"")
				KillWaves $PossiblyQuotePath(tempPathStr,":") ; tempPathStr = ReplaceString("_nc4_non_coord_",tempPathStr,"")
				Note/K $PossiblyQuotePath(tempPathStr,":"), ReplaceString("_nc4_non_coord_",note($PossiblyQuotePath(tempPathStr,":")),"")
			endif
			tmpVarLstStr+=ReplaceString("time0",ParseFilePath(0,tempPathStr,":",1,0),timeWvStrConstant)+";"
		endfor
				
		// add file information
		QLA_set_att(file_info,"variables",num2istr(ii)+"; // int")
		QLA_set_att(file_info,"names","\""+tmpVarLstStr+"\";")
					
		// Check for 'time/Time/TIME' variables
		HDF5ListGroup /F /R /TYPE=3 fileID, "/"
		for (i=0;i<ItemsInList(S_dataFolderPaths);i+=1)
			Wave/Z timeWv= $StringFromList(i,S_dataFolderPaths)+"time0"
			if(WaveExists(timeWv))
				SetScale/P x 0,1,QLA_get_att(TimeWV,"units"), TimeWv
				SetScale d 0,0,"dat", TimeWV
				// Rename 'time' to whatever constant is defined
				Duplicate/O timeWv, $StringFromList(i,S_dataFolderPaths)+timeWvStrConstant ; KillWaves/Z timeWv
			endif
		endfor
		
		// Close the file
//		HDF5DumpErrors/Q 	// Retrieve collected error information
//		Print S_HDF5ErrorInfo
		HDF5CloseFile/Z fileID
			
	else // netCDF-3 fall-back
			
		// Open the netCDF-3				
		NC_OpenFile/Z/PATH=netCDFpath fileID as fNameStr  
		
		// Handle errors
		if (V_flag != 0) // User chose an invalid netCDF, already established it wasnt NC4/H5 
			DoAlert/T="netCDF Loader", 0, "The file:\r\r \'"+fNameStr+"\'\r\r . . . is not a valid netCDF-3/4 and was not loaded"
			return "" 
		endif
		
		// Load all global attributes
		NC_LoadData/O/ALL/A/T=1 fileID	
		
		// sort attributes alphanumerically
		#ifdef AlphaNumSort
			S_names = SortList(S_names,";",16)
		#endif		 			

		// Get global attributes
		Make/O/T/N=(0) $"global_attributes"/Wave=infoWv
		for (i=0;i<ItemsInList(S_names);i+=1)
			AttStr=ReplaceString("global_",StringFromList(i,S_names),"")
			// string
			if ((WaveType($StringFromList(i,S_names))==0) && (numpnts($StringFromList(i,S_names))!=0)) 
				Wave/Z/T AttStrWv = $StringFromList(i,S_names)
				QLA_set_att(infoWv,attStr,"\""+AttStrWv[Inf]+"\";")
			// numeric
			elseif  ((WaveType($StringFromList(i,S_names))!=0) && (numpnts($StringFromList(i,S_names))!=0))  
				Wave/Z	AttVarWv = $StringFromList(i,S_names)
				QLA_set_att(infoWv,attStr,cprintf(AttVarWv)+"; // "+typeToinfo(WaveType(AttVarWv)))
			// empty
			else
				KillWaves/Z $StringFromList(i,S_names) // kill here as never declared
				QLA_set_att(infoWv,attStr,"\"n/a\";") 
			endif
			KillWaves/Z AttStrWv, AttVarWv
		
		// end of global metadata loop
		endfor 
		
		// make the attributes title 
		Note/K infoWv, "// global attributes:\r"+note(infoWv)
		
		// Kill the info Wave if it is empty
		if (ItemsInList(QLA_inq_att(infoWv)) == 0)
			KillWaves/Z infoWv
		endif
		
		// get dimensions
		Make/O/T/N=(0) $"dimensions"/Wave=dimWv
		Note dimWv,getNCdimensionStr(fileID)
			
		// list all variables in file
		NC_ListObjects /T="var" fileID 
		String varNameStr=S_names
		
		// Get variable & attributes	
		for (i=0;i<ItemsInList(varNameStr);i+=1)
			varStr = StringFromList(i,varNameStr)
			
			// Load variable
			NC_LoadData /O /U=3 /T=0 fileID, varStr
			
			// check for time/Time/TIME variables
			if	(CmpStr(StringFromList(i,varNameStr),"time")==0)		
				Wave/Z dataWv=$StringFromList(i,ReplaceString("ime",varNameStr,"ime0")) 
			else
				Wave/Z dataWv=$varStr
			endif
			
			// check to see if data was actually loaded 
			if (!WaveExists(dataWv))
				continue
			endif
			
			// kill the wavenote set the short name
			Note/K dataWv ;	QLA_set_att(dataWv,"short_name","\""+varStr+"\";")
			
			// Load attributes for variable
			NC_LoadData /O /All /A /T=1 fileID, varStr
						
			// sort attributes alphanumerically
			#ifdef AlphaNumSort
				S_names = SortList(S_names,";",16)
			#endif		 			

			// get the attributes
			for(ii=0;ii<ItemsInList(S_names);ii+=1)
				attStr = ReplaceString(varStr+"_",StringFromList(ii,S_names),"")
				// string
				if ((WaveType($StringFromList(ii,S_names)) == 0) && (numpnts($StringFromList(ii,S_names)) != 0)) 
					Wave/Z/T AttStrWv = $StringFromList(ii,S_names)
				  	QLA_set_att(dataWv,attStr,"\""+AttStrWv[Inf]+"\";")
				 // numeric
				elseif ((WaveType($StringFromList(ii,S_names)) !=0 ) && (numpnts($StringFromList(ii,S_names)) != 0))  
				  	Wave/Z	AttVarWv = $StringFromList(ii,S_names)
				  	QLA_set_att(dataWv,attStr, cprintf(AttVarWv)+"; // "+typeToinfo(WaveType(AttVarWv)))
  				// empty
  				else 
  					KillWaves/Z $StringFromList(ii,S_names) // kill here as never declared
  					QLA_set_att(dataWv,attStr,"n/a") 
				endif	
				KillWaves/Z AttStrWv, AttVarWv
								 
			 	// transpose 2D datasets if required
				if (transpose && (WaveDims(dataWv) > 1))
				 	MatrixTranspose dataWv
				endif	
			
				// apply metadata conventions to wave
				unpackNC(dataWv,mode)
			endfor
				
			// get dimension information
			NC_Inquire/Z/Q fileID, varStr 
			setDependencies(dataWv,S_names)
			
			setDim(dimWv,StringFromList(1,S_names),StringFromList(1,S_dims))

		// end of variable metadata loop
		endfor
			
		// Deal with Time which is illegal in Igor
		if (Exists("Time0"))
			Wave Time0
			SetScale/P x 0,1,QLA_get_att(Time0,"units"), Time0
			SetScale d 0,0,"dat", Time0
			Duplicate/O Time0 $timeWvStrConstant ; KillWaves/Z Time0
		endif

//		NC_DumpErrors/C	          // Retrieve collected error information
// 		print S_errorInfo
		NC_CloseFile fileID
		
		// add file information
		QLA_set_att(file_info,"type","\"netCDF-3\";")
		QLA_set_att(file_info,"variables",num2istr(i)+"; // int")
		QLA_set_att(file_info,"names","\""+ReplaceString("time",varNameStr,timeWvStrConstant)+"\";")

	endif	// end of netCDF-3 switch

	return fNameStr
end
//___________________________________________________________________________________________________
// return the dimesnion names from the CDL syntax
Function/S QLA_get_dims2(Wave dataWv)
	String rtnStr="", dimStr,dataStr,matchStr
	
	dataStr = ParseFilePath(0,note(dataWV),";",0,0)
	matchStr=".*"+NameOfWave(dataWv)+"\("
	
	variable i
	for (i=0;i<WaveDims(dataWV);i+=1)
		SplitString/E=(matchStr+"(.*?)=.*") dataStr,dimStr
		matchStr+=dimStr+"=.*, "
		rtnStr+=dimStr+";"
	endfor
	return rtnStr
end 
//___________________________________________________________________________________________________	
// return the dimesnion names	 from the labels
Function/S QLA_get_dims(Wave dataWv)
	String rtnStr=""
	
	variable i
	for (i=0;i<WaveDims(dataWV);i+=1)
		rtnStr+=GetDimLabel(dataWV,i,-1)+";"
	endfor
	return rtnStr
end 
//___________________________________________________________________________________________________	
// set or modify the named attribute
Function QLA_set_att(Wave Wv, String attKey, String attVal)

	if (strlen(attVal)==0)
		Note/K Wv, RemoveByKey("\t:"+attKey,note(Wv)," = ","\r")
	else
		Note/K Wv, ReplaceStringByKey("\t:"+attKey,note(Wv),attVal," = ","\r")
	endif
end
//___________________________________________________________________________________________________	
// get the named attbribute as a string
Function/S QLA_get_att(Wave Wv,String attKey)

	String rtnStr
	SplitString/E=("(.*?);.*") ReplaceString("\"",StringByKey("\t:"+attKey,Note('Wv')," = ","\r"),""),rtnStr
	return rtnStr
end
//___________________________________________________________________________________________________	
// list the attributes
Function/S QLA_inq_att(Wave Wv)

	Variable i
	String attStr,valStr,lstStr,rtnStr="" 
	for (i=0;i<ItemsInList(note(Wv),"\t:");i+=1)
		lstStr = ReplaceSTring("\"",StringFromList(i,note(Wv),"\t:"),"")
		SplitString/E=("(.*?) = (.*?)\r") lstStr,attStr,valStr 
		if (strlen(attStr)>0)
			rtnStr=AddListItem(AttStr,rtnStr)
		endif
	endfor
	return rtnStr
end
//___________________________________________________________________________________________________		
// apply unpack preferences 
Static Function unpackNC(Wave dataWv,int mode)

	// list of all the attributes associated with dataWv
	String attsLst = QLA_inq_att(dataWv)
	
	Variable fillVal ; String fillStr
	
	// get a fill value 
	if (GrepString(attsLst,"_FillValue") && (WaveType(dataWv,1)==1) && (WaveType(dataWv)<8) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"_FillValue"),",",WaveType(dataWv))
		fillVal = attWv[inf]
	elseif (GrepString(attsLst,"missing_value") && (WaveType(dataWv,1)==1) && (WaveType(dataWv)<8) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"missing_value"),",",WaveType(dataWv))
		fillVal = attWv[inf]
	endif
		
	// units
	if (GrepString(attsLst,"units"))
		SetScale d 0,0, QLA_get_att(dataWv,"units"), dataWv
	endif
	
	// scale_factor - before offset
	if ((mode<2 || mode==3) && (GrepString(attsLst,"scale_factor")) && (WaveType(dataWv,1)==1) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"scale_factor"),",",WaveType(dataWv))
		if (attWv[inf] != 1)	// waste of time multiplying every point by 1
			Redimension/S dataWv
			dataWv *= attWv[inf]
		endif
	endif

	// add_offset - after scaling
	if ((mode<2 || mode==3) && (GrepString(attsLst,"add_offset")) && (WaveType(dataWv,1)==1) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"add_offset"),",",WaveType(dataWv))
		if (attWv[inf] != 0)	// waste of time adding zero to every point
			dataWv += attWv[inf]
		endif
	endif				
	
	// valid_range	
	if ((mode<3) && (GrepString(attsLst,"valid_range")) && (WaveType(dataWv,1)==1) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_range"),",",WaveType(dataWv))						
		dataWv = (attWv[0] > dataWv[p][q][r][s])||(dataWv[p][q][r][s] > attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif						
	
	// valid_max
	if ((mode<3) && (GrepString(attsLst,"valid_max")) && (WaveType(dataWv,1)==1) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_max"),",",WaveType(dataWv))
		dataWv = (dataWv[p][q][r][s] > attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif
	
	// valid_min
	if ((mode<3) && (GrepString(attsLst,"valid_min")) && (WaveType(dataWv,1)==1) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"valid_min"),",",WaveType(dataWv))
		dataWv = (dataWv[p][q][r][s] < attWv[Inf]) ? fillVal : dataWv[p][q][r][s]
	endif
	
	// _FillValue - set to NaN in floating point waves
	if ((mode==0) && (GrepString(attsLst,"_FillValue")) && (WaveType(dataWv,1)==1) && (WaveType(dataWv)<8) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"_FillValue"),",",WaveType(dataWv))
		dataWv = (dataWv[p][q][r][s]== attWv[inf]) ? NaN : dataWv[p][q][r][s]
	elseif ((mode==0) && (GrepString(attsLst,"_FillValue")) && (WaveType(dataWv,1)==2) && (numpnts(dataWv)>0))
		Wave/T textWv = dataWv ; fillStr = QLA_get_att(dataWv,"_FillValue")
		textWv = ReplaceString(fillStr,textWv,"") // is this even allowed / are _FillValues allowed in string/char data?
	endif		 			
	
	// missing_value	 - set to NaN in floating point waves			 			
	if ((mode==0) && (GrepString(attsLst,"missing_value")) && (WaveType(dataWv,1)==1) && (WaveType(dataWv)<8) && (numpnts(dataWv)>0))
		Wave attWv = varLst2wave(QLA_get_att(dataWv,"missing_value"),",",WaveType(dataWv))
		dataWv = (dataWv[p][q][r][s] == attWv[inf]) ? NaN : dataWv[p][q][r][s]
	elseif ((mode==0) && (GrepString(attsLst,"missing_value")) && (WaveType(dataWv,1)==2) && (numpnts(dataWv)>0))
		Wave/T textWv = dataWv ; fillStr = QLA_get_att(dataWv,"missing_value")
		textWv = ReplaceString(fillStr,textWv,"") // is this even allowed / are missing_values allowed in string/char data?
	endif		 	
end
//___________________________________________________________________________________________________	
// function to get a list of netCDF/HDF5 dimensions from DIMENSION_LIST attribute
Static Function/S getHDF5Dimensions(Variable groupID, String WvStr)

	Variable i ; String rtnStr=""
	
	HDF5ListAttributes /Type=2 groupID, WvStr
	
	// variabe has dimensions attached
	if (GrepString(S_HDF5ListAttributes,"DIMENSION_LIST"))
	
		// load the dimension list attribute 
		HDF5LoadData /O /A="DIMENSION_LIST" /TYPE=2 /N=temp_DIMENSION_LIST /Q /Z groupID, WvStr
		
		// get a list of all possible dimension list waves
		String dimWvLst = WaveList("temp_DIMENSION_LIST*",";","")
		if (V_flag==0)	
			for (i=0;i<ItemsInList(dimWvLst);i+=1)
				Wave/T temp_DimLstWv = $StringFRomLIst(i,dimWvLst)
				rtnStr+=ParseFilePath(0,temp_DimLstWV[0],"/",1,0)+";"
				KillWaves temp_DimLstWv
			endfor
		endif
	endif
	
	// is a dimension
	HDF5DimensionScale dataset={groupID,WvStr}, isScale
	if (V_Value)
		rtnStr=ParseFilePath(0,WvStr,"/",1,0)+";"
	endif
	
	return rtnStr
end	
//___________________________________________________________________________________________________	
// check if the dataset is a dimension even if it isn't attached to a variable
Static Function/S isHDF5Dimension(Variable groupID,String WvStr)

	Variable dummy ; String rtnStr=""
	
	HDF5DimensionScale dataset={groupID,WvStr}, isScale

	if (V_Value)
 
 		STRUCT HDF5DataInfo di	
		InitHDF5DataInfo(di)
		dummy = HDF5DatasetInfo(groupID, WvStr, 0, di)
		WvStr = ParseFilePath(0,WvStr,"/",1,0)
		
		if (di.maxdims[0] == -1)
			rtnStr="\t"+Wvstr+" = UNLIMITED;   // ("+num2istr(di.dims[0])+" currently)\r"
		else 
			rtnStr="\t"+Wvstr+" = "+num2istr(di.dims[0])+";\r"
		endif
	endif

	return rtnStr
end
//___________________________________________________________________________________________________	
// function to return a formatted string with netCDF dependencies in it 
// also set dim labels with those dependencies if dimensions has >0 size
Static Function setDependencies(Wave Wv, String dimLstStr)

	Variable i ; String dimStr=""
	String rtnStr = TypeToInfo(WaveType(Wv))+" "+QLA_get_att(Wv,"short_name")+"("
	for (i=0;i<WaveDims(Wv);i+=1)
		if (Dimsize(wv,i)>0 && strlen(dimLstStr)>0) // cannot set a dim label on a 0 size dimension
			SetDimLabel i, -1, $StringFromList(i,dimLstStr), Wv
			dimStr=StringFromList(i,dimLstStr)+"="
		endif
		rtnStr+=dimStr+num2iStr(DimSize(Wv,i))+", "
	endfor 
		
	Note/K Wv, RemoveEnding(rtnStr,", ")+");\r"+note(Wv)
end
//___________________________________________________________________________________________________	
// Function to determine the type of netCDF variable
Static Function/S TypeToInfo(type)
	Variable type
	
	String info = ""

	switch(type)
		case 0: // added CR 04/20
			info="string"
			break
		case 1: // added CR 05/21
			info="char"
			break
		case 2:
			info="float"
			break
		case 4:
			info="double"
			break
		case 8:
			info="byte"
			break
		case 16:
			info="short"
			break
		case 32:
			info="int" // long
			break
		case 72:  // added CR 04/20
			info="ubyte"
			break
		case 80:  // added CR 04/20
			info="ushort"
			break
		case 96:  // added CR 04/20
			info="uint"
			break
		case 128:  // added CR 02/21
			info="int64"
			break
		case 192:  // added CR 02/21
			info="uint64"
			break
		default:
			info=""
	endswitch

	return info
End
//___________________________________________________________________________________________________  
// return a path possibly quoted for liberal names
Static Function/S PossiblyQuotePath(String pathStr, String sepStr)
	
	Variable i ; String rtnStr=""

	pathStr+=sepStr
	
	for (i=0;i<ItemsInList(pathStr,sepStr);i+=1)
		rtnStr+=PossiblyQuoteName(StringFromList(i,pathStr,sepStr))+sepStr
	endfor
	
	return RemoveEnding(rtnStr,sepStr)
end
//___________________________________________________________________________________________________	
// convert a stringlist into a wave of given type in formatstring 
Static Function/WAVE varLst2wave(String varLstStr,String sepStr,int type)

	Variable n = ItemsInList(varLstStr, sepStr)
	Make/O/N=(n)/Y=(type)/FREE outWv = str2num(StringFromList(p,varLstStr, sepStr))

	return outWv
end
//___________________________________________________________________________________________________	
// control printf or maybe chris printf!
Static Function/S cprintf(Wave Wv)

	String rtnStr="",cmdStr ; int i,j=numpnts(wv),type = WaveType(Wv),k,l
	
	for (i=0;i<j;i+=1)

		if (type == 2)
			[k,l]=getPrecision(Wv[i],7) // single can represent ~ 7 dp
			sprintf cmdStr,"%#*.*f, " k,l,wv[i]
			rtnStr+=cmdStr
		elseif (type == 4)
			[k,l]=getPrecision(Wv[i],15) // double can represent ~ 15 dp
			sprintf cmdStr,"%#*.*f, " k,l,wv[i]
			rtnStr+=cmdStr
		elseif (type > 4)
			sprintf cmdStr, "%d, ", Wv[i]
			rtnStr+=cmdStr
		endif
	endfor 

	return RemoveEnding(rtnStr,", ")
end
//___________________________________________________________________________________________________	
// get the scale and precision of a number, i.e. number of decimal places 
// number of DP based on how large the integer part is to avoid overstated precision
Static Function [int integers,int decimals] getPrecision(Variable num, int exponent)
   
   double threshold
   
   integers = ceil(log(abs(num)))   
	threshold = 5*10^-(exponent-integers+1) 
	num-=floor(num)

	do
		num *= 10.0;
		threshold *= 10.0;
		decimals += 1
	while (abs(num - round(num)) > threshold) 

	return [integers,decimals] 
end	
//___________________________________________________________________________________________________	
// function called if the netCDF XOP isn't found
Static Function QLA_error()
	DoAlert/T="netCDF loader" 0,"Please install the NetCDF XOP and restart Igor"
	
	// try and give some help for the netCDF XOP
	DisplayHelpTopic/Z "Installing NetCDF XOP"
	if (V_flag !=0)
		// help for XOPs in general
		DisplayHelpTopic/Z "File Loader XOPs"
	endif
	Abort
end
//___________________________________________________________________________________________________
// get the dimension imformation from a netCDF-3
Static Function/S getNCdimensionStr(Variable fileID)

	String rtnStr="dimensions:\r", varStr,dimStr ; int i
	
	// get unlimited dimensions
	NC_ListObjects /Z /Q /T="udim;" fileID 
	if (V_flag == 0)
		dimStr = S_names
		if (ItemsInList(dimStr)>0)
			varStr = StringFromList(0,dimStr) // only 1 unlimited dimension is allowed in netCDF classic
			NC_Inquire /Z fileID, varStr
			if (V_flag == 0)
				rtnStr+="\t"+varStr+" = UNLIMITED;   // ("+RemoveEnding(S_dims)+" currently)\r"
			else 
				rtnStr+="\t"+varStr+" = UNKNOWN;\r"
			endif
		endif 
	endif
	
	// get finite dimensions
	NC_ListObjects /Z /Q /T="dim;" fileID 
	if (V_flag == 0)
		dimStr = S_names
		for (i=0;i<ItemsInList(dimStr);i+=1)
			varStr = StringFromList(i,dimStr)
			NC_Inquire /Z fileID, varStr
			if (V_flag == 0)
				rtnStr+="\t"+varStr+" = "+S_dims+"\r"
			else
				rtnStr+="\t"+varStr+" = UNKNOWN;\r"
			endif
		endfor
	endif

	return rtnStr
end
//___________________________________________________________________________________________________
// get dimesnions size
Static Function/S getDim(Wave Wv,String dimStr)

	return RemoveEnding(StringByKey("\t"+dimStr,Note('Wv')," = ","\r"),";")
end
//___________________________________________________________________________________________________
// set the size of a dimension of unknown size i.e. a dimension but not a variable
Static Function setDim(Wave Wv,String dimStr,String dimLen)

	String noteStr = ReplaceString(")\r",Note('Wv'),");\r")
	if (GrepString(getDim(Wv,dimStr),"UNKNOWN"))
		Note/K Wv, ReplaceString(");\r",ReplaceStringByKey("\t"+dimStr,noteStr,dimLen," = ",";\r"),")\r")
	endif
end
//___________________________________________________________________________________________________
// list the dimesnions 
Static Function/S listDims(Wave Wv)

	Variable i
	String dimStr,lstStr,rtnStr="" 
	for (i=0;i<ItemsInList(note(Wv),"\t");i+=1)
		lstStr = StringFromList(i,note(Wv),"\t")
		SplitString/E=("(.*?) = .*\r") lstStr,dimStr 
		if (strlen(dimStr)>0)
			rtnStr+=dimStr+";"
		endif
	endfor
	
	return rtnStr
end
//___________________________________________________________________________________________________
// try to determine if a text wave is really char
Static Function isChar(Wave Wv)

	int type = WaveType(Wv)
	
	// is numeric 
	if (type > 0)
		return 0
	endif 
	
	// check string length 
	int len = getMaxStrLen(Wv)
	
	if (len == 1)
		return 1
	endif		
		
	return 0
End
//___________________________________________________________________________________________________	
// get maximum string length in a text wave
Static Function getMaxStrLen(Wave/T Wv)

	int maxlen,len,i
	do 
		len = strlen(Wv[i])
		maxLen = (len > maxLen) ? len : maxLen
		i++
	while (i < numpnts(Wv))
	
	return maxLen
End
//___________________________________________________________________________________________________	
